# run this command
import flask
import dash
import dash_html_components as html


server = flask.Flask(__name__)

@server.route('/')
def index():
    return 'Flask root.'

from foresee.webapp.dash_app import app

if __name__ == '__main__':
    app.run_server()
    


"""
utility script for optimization techniques
"""


import warnings
warnings.filterwarnings('ignore')

from hyperopt import hp, fmin, tpe, Trials

# local methods
from foresee.models import ewm
from foresee.models import holt_winters
from foresee.models import fft
from foresee.models import sarimax

from foresee.models import models_util

def _model_wfa(p, ts_train, ts_test, model, freq, epsilon):
    
    fcst_len = len(ts_test)
    
    if model == 'ewm_model':
        
        fittedvalues, yhat, err  = ewm.ewm_fit_forecast(ts_train, fcst_len, p)
        
        if err is None:
            wfa = models_util.compute_wfa(ts_test.values, yhat.values, epsilon)
            
        else:
            wfa = 0
        
    
    elif model == 'holt_winters':
        
        fittedvalues, yhat, err  = holt_winters.holt_winters_fit_forecast(ts_train, fcst_len, freq, p)
        
        if err is None:
            wfa = models_util.compute_wfa(ts_test.values, yhat.values, epsilon)
            
        else:
            wfa = 0
        
    elif model == 'fft':
        
        fittedvalues, yhat, err  = fft.fft_fit_forecast(ts_train, fcst_len, p)
        
        if err is None:
            wfa = models_util.compute_wfa(ts_test.values, yhat.values, epsilon)
            
        else:
            wfa = 0
            
    elif model == 'sarimax':
        
        fittedvalues, yhat, err  = sarimax.sarimax_fit_forecast(ts_train, fcst_len, freq, p)
        
        if err is None:
            wfa = models_util.compute_wfa(ts_test.values, yhat.values, epsilon)
            
        else:
            wfa = 0
            print(err)
            
    else:
        wfa = 0
        
    return 1 - wfa


def tune(train_fact, test_fact, fcst_len, model, model_param_space, freq, epsilon):
    
    ts_train = train_fact['y']
    ts_test = test_fact['y']
    
    best, trials = optimize_param(ts_train, ts_test, fcst_len, model, model_param_space, freq, epsilon)
    
    return best, trials


def optimize_param(ts_train, ts_test, fcst_len, model, model_param_space, freq, epsilon):
    
    # TODO: fill out function description
    
    if model == 'ewm_model':
        
        try:
            span_lb = 2
            span_ub = model_param_space
            
            space = hp.quniform('span', span_lb, span_ub, 1)
            
            f_obj = lambda p: _model_wfa(p, ts_train, ts_test, model, freq, epsilon)
            
            trials = Trials()
            
            best = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=100, show_progressbar=False, verbose=False)
            
            best = best['span']
            
        except Exception as e:
            best = str(e)
            
    elif model == 'holt_winters':
        
        try:
            space = {
                        's_level': hp.uniform('s_level', 0, 1),
                        's_slope': hp.uniform('s_slope', 0, 1),
                        's_season': hp.uniform('s_season', 0, 1),
                        'd_slope': hp.uniform('d_slope', 0, 1)
                    }
            
            f_obj = lambda p: _model_wfa(p, ts_train, ts_test, model, freq, epsilon)
            
            trials = Trials()
            
            best = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=200, show_progressbar=False, verbose=False)
            
        except Exception as e:
            best = str(e)
        
    elif model == 'fft':
        
        try:
            nh_lb = 2
            nh_ub = model_param_space
            
            space = hp.quniform('nh', nh_lb, nh_ub, 1)
            
            f_obj = lambda p: _model_wfa(int(p), ts_train, ts_test, model, freq, epsilon)
            
            trials = Trials()
            
            best = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=100, show_progressbar=False, verbose=False)
            
            best = int(best['nh'])
            
        except Exception as e:
            best = str(e)
            
        
    elif model == 'sarimax':
        
        try:
            max_p = model_param_space['p']
            max_d = model_param_space['d']
            max_q = model_param_space['q']
            max_cp = model_param_space['cp']
            max_cd = model_param_space['cd']
            max_cq = model_param_space['cq']
            
            space = {
                        'p': hp.randint('p', max_p),
                        'd': hp.randint('d', max_d),
                        'q': hp.randint('q', max_q),
                        'cp': hp.randint('cp', max_cp),
                        'cd': hp.randint('cd', max_cd),
                        'cq': hp.randint('cq', max_cq)
                    }
            
            f_obj = lambda p: _model_wfa(p, ts_train, ts_test, model, freq, epsilon)
            
            trials = Trials()
            
            best = fmin(f_obj, space, algo=tpe.suggest, trials=trials, max_evals=200, show_progressbar=False, verbose=False)
            
        except Exception as e:
            best = str(e)
            print(best)
        
    else:
        raise ValueError(model)
        
    return best, trials
    




"""
exponentially weighted moving average
"""

import numpy as np
import pandas as pd

# local module
from foresee.models import models_util
from foresee.models import param_optimizer


def ewm_fit_forecast(ts, fcst_len, span):
    
    try:
        
        ewm_model = pd.Series(ts).ewm(span=span)
        ewm_fittedvalues = ewm_model.mean()
        
        ewm_forecast = np.empty(fcst_len)
        ewm_forecast.fill(ewm_fittedvalues.iloc[-1])
        ewm_forecast = pd.Series(ewm_forecast)
        err = None

    except Exception as e:
        
        ewm_fittedvalues = None
        ewm_forecast = None
        err = str(e)
        
    return ewm_fittedvalues, ewm_forecast, err


def fit_ewm(data_dict, freq, fcst_len, model_params, run_type, tune, epsilon):
    
    model = 'ewm_model'
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    
    # no model competition
    if run_type in ['all_models']:
        
        span = 5
        ewm_fitted_values, ewm_forecast, err = ewm_fit_forecast(
                                                                    ts = complete_fact['y'],
                                                                    fcst_len = fcst_len,
                                                                    span = span,
                                                                )
        
        if err is None:
            fit_fcst_fact['ewm_model_forecast'] = ewm_fitted_values.append(ewm_forecast).values
            
        else:
            fit_fcst_fact['ewm_model_forecast'] = 0
            
        args['err'] = err
        args['span'] = span
            
            
    # with model completition            
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            model_param_space = model_params[model]['span']
            span, trials = param_optimizer.tune(train_fact, test_fact, fcst_len, model, model_param_space, freq, epsilon)
            args['trials'] = trials
            
        else:
            span = 5
            
        training_fitted_values, training_forecast, training_err = ewm_fit_forecast(
                                                                                    ts = train_fact['y'],
                                                                                    fcst_len = len(test_fact) ,
                                                                                    span = span
                                                                                )
        
        complete_fitted_values, complete_forecast, complete_err = ewm_fit_forecast(
                                                                                    ts = complete_fact['y'],
                                                                                    fcst_len = fcst_len,
                                                                                    span = span
                                                                                )
        
        if training_err is None and complete_err is None:
            ewm_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = training_forecast.values,
                                    epsilon = epsilon,
                                )
            ewm_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['ewm_model_forecast'] = ewm_fit_fcst.values
            fit_fcst_fact['ewm_model_wfa'] = ewm_wfa
            
        else:
            ewm_wfa = -1
            fit_fcst_fact['ewm_model_forecast'] = 0
            fit_fcst_fact['ewm_model_wfa'] = ewm_wfa
            
        args['err'] = (training_err, complete_err)
        args['wfa'] = ewm_wfa
        args['span'] = span
            
            
    return fit_fcst_fact, args



"""
collection of utility functions used in models section
"""

import numpy as np
import pandas as pd

def compute_wfa(y, yhat, epsilon):
    
    """
    weighted forecast accuracy
    
    TODO: update error checking 
    """
    
    try:
        abs_err = compute_abs_err(y, yhat)
        wfa = max(0, 1-(2*sum(abs_err))/(sum(y) + sum(yhat) + epsilon))
        
    except Exception as e:
        wfa = 0
        err = str(e)
        print(err)
        
    return wfa
    
    
def compute_abs_err(y, yhat):
    
    #TODO: add more checks, e.g. data-type, same length, ...
    
    if any(np.isnan(y)) or any(np.isnan(yhat)):
        raise ValueError('error call from abs_err func: y or yhat has null values')
        
    if len(y) == 0 or len(yhat)  == 0:
        raise ValueError('error call from abs_err func: y or yhat is empty')
        
    return abs(y - yhat)


# failed attempt to extend dataframe with date-time column, too many unknowns (TODO: try again)
def make_future_dataframe():
    
    df = pd.DataFrame()
    
    dates = pd.date_range()
    
    return df






"""
Holt-Winters
"""

import warnings
warnings.filterwarnings('ignore')

import numpy as np
import pandas as pd
from statsmodels.tsa.holtwinters import ExponentialSmoothing

# local module
from foresee.models import models_util
from foresee.models import param_optimizer

def holt_winters_fit_forecast(ts, fcst_len, freq, params):
    
    try:
        hw_model = ExponentialSmoothing(
                                            endog = ts,
                                            trend = 'add',
                                            damped = True,
                                            seasonal = 'add',
                                            seasonal_periods = freq
                                          )
        
        if params is None:
            hw_model = hw_model.fit(optimized = True)
            
        else:
            s_level, s_slope, s_season, d_slope = (params['s_level'], params['s_slope'], params['s_season'], params['d_slope'])
            hw_model = hw_model.fit(
                                                      smoothing_level = s_level,
                                                      smoothing_slope = s_slope,
                                                      smoothing_seasonal = s_season,
                                                      damping_slope = d_slope,
                                                      optimized = True,
                                    )
            
        hw_fittedvalues = hw_model.fittedvalues    

        hw_forecast = hw_model.predict(
                                            start = len(ts),
                                            end =   len(ts) + fcst_len - 1
                                        )
        err = None
        
    except Exception as e:
        hw_fittedvalues = None
        hw_forecast = None
        err = str(e)
    
    
    return hw_fittedvalues, hw_forecast, err


def fit_holt_winters(data_dict, freq, fcst_len, model_params, run_type, tune, epsilon):
    
    model = 'holt_winters'
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()

    # no model competition    
    if run_type in ['all_models']:
        
        hw_fitted_values, hw_forecast, err = holt_winters_fit_forecast(
                                                                           ts = complete_fact['y'],
                                                                           fcst_len = fcst_len,
                                                                           freq = freq,
                                                                           params = None
                                                                      )
        if err is None:
            fit_fcst_fact['holt_winters_forecast'] = hw_fitted_values.append(hw_forecast).values
            
        else:
            fit_fcst_fact['holt_winters_forecast'] = 0
            
        args['err'] = err
    
    
    # with model completition            
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            
            model_param_space = model_params[model]
            opt_params, trials = param_optimizer.tune(train_fact, test_fact, fcst_len, model, model_param_space, freq, epsilon)
            
            args['trials'] = trials
            
        else:
            opt_params = None
            
        training_fitted_values, training_forecast, training_err = holt_winters_fit_forecast(
                                                                                                   ts = train_fact['y'],
                                                                                                   fcst_len = len(test_fact),
                                                                                                   freq = freq,
                                                                                                   params = opt_params,
                                                                                         )
        
        complete_fitted_values, complete_forecast, complete_err = holt_winters_fit_forecast(
                                                                                                   ts = complete_fact['y'],
                                                                                                   fcst_len = fcst_len,
                                                                                                   freq = freq,
                                                                                                   params = opt_params,
                                                                                         )
        
        if training_err is None and complete_err is None:
            hw_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = training_forecast.values,
                                    epsilon = epsilon,
                                )
            hw_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['holt_winters_forecast'] = hw_fit_fcst.values
            fit_fcst_fact['holt_winters_wfa'] = hw_wfa
            
        else:
            hw_wfa = -1
            fit_fcst_fact['holt_winters_forecast'] = 0
            fit_fcst_fact['holt_winters_wfa'] = -1
            
        args['err'] = (training_err, complete_err)
        args['wfa'] = hw_wfa
        args['params'] = opt_params
        
        
    return fit_fcst_fact, args







# -*- coding: utf-8 -*-
"""
local utility functions 
"""

import os
import json
import pandas as pd
import numpy as np
from io import StringIO
from importlib_resources import files

def read_json(file_name):
    
    json_string = files('foresee.params').joinpath(file_name).read_text()
    
    return json.loads(json_string.replace('\n', ' ').replace('\t', ' '))


def read_csv(file_name):
    
    data_txt = files('foresee.data').joinpath(file_name).read_text()
    
    return pd.read_csv(StringIO(data_txt))



"""
sarimax from statsmodels
"""


import warnings
warnings.filterwarnings('ignore')

import numpy as np
import pandas as pd
import statsmodels.api

# local module
from foresee.models import models_util
from foresee.models import param_optimizer


def sarimax_fit_forecast(ts, fcst_len, freq, params):
    
    try:
        if params is None:
            sarimax_model = statsmodels.api.tsa.statespace.SARIMAX(
                                                                        endog = ts,
                                                                        enforce_stationarity = False,
                                                                        enforce_invertibility = False,
                                                                   ).fit(disp=0)
        else:
            order = (params['p'], params['d'], params['q'])
            seasonal_order = (params['cp'], params['cd'], params['cq'], freq)
            
            sarimax_model = statsmodels.api.tsa.statespace.SARIMAX(
                                                                        endog = ts,
                                                                        order = order,
                                                                        seasonal_order = seasonal_order,
                                                                        enforce_stationarity = True,
                                                                        enforce_invertibility = True,
                                                                   ).fit(disp=0)
            
        
        sarimax_fittedvalues = sarimax_model.fittedvalues
        sarimax_forecast = sarimax_model.predict(
                                                    start = len(ts),
                                                    end = len(ts) + fcst_len - 1,
                                                 )
        err = None
        
    except Exception as e:
        
        sarimax_model = None
        sarimax_fittedvalues = None
        sarimax_forecast = None
        err = str(e)
        
    return sarimax_fittedvalues, sarimax_forecast, err


def fit_sarimax(data_dict, freq, fcst_len, model_params, run_type, tune, epsilon):
    
    model = 'sarimax'
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    
    # no model competition
    if run_type in ['all_models']:
        
        sarimax_fitted_values, sarimax_forecast, err = sarimax_fit_forecast(
                                                                            ts = complete_fact['y'],
                                                                            fcst_len = fcst_len,
                                                                            freq = freq,
                                                                            params = None,
                                                                        )
        
        if err is None:
            fit_fcst_fact['sarimax_forecast'] = sarimax_fitted_values.append(sarimax_forecast).values
            
        else:
            fit_fcst_fact['sarimax_forecast'] = 0
            
        args['err'] = err
    
    # with model completition            
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            model_param_space = model_params[model]
            params, trials = param_optimizer.tune(train_fact, test_fact, fcst_len, model, model_param_space, freq, epsilon)
            args['trials'] = trials
            
        else:
            params = None
            
        training_fitted_values, training_forecast, training_err = sarimax_fit_forecast(
                                                                                    ts = train_fact['y'],
                                                                                    fcst_len = len(test_fact),
                                                                                    freq = freq,
                                                                                    params = params,
                                                                                )
        
        complete_fitted_values, complete_forecast, complete_err = sarimax_fit_forecast(
                                                                                    ts = complete_fact['y'],
                                                                                    fcst_len = fcst_len,
                                                                                    freq = freq,
                                                                                    params = params,
                                                                                )
        
#        if 'enforce_stationarity' in complete_err or '' in complete_err:
#            
#            # run without enforcement
#        
#        #TODO: if failed with 'enforce_stationarity' or 'enforce_invertibility' at complete fcst, set these to false
        
        if training_err is None and complete_err is None:
            sarimax_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = training_forecast.values,
                                    epsilon = epsilon,
                                )
            sarimax_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['sarimax_forecast'] = sarimax_fit_fcst.values
            fit_fcst_fact['sarimax_wfa'] = sarimax_wfa
            
        else:
            sarimax_wfa = -1
            fit_fcst_fact['sarimax_forecast'] = 0
            fit_fcst_fact['sarimax_wfa'] = -1
            
        args['err'] = (training_err, complete_err)
        args['wfa'] = sarimax_wfa
        args['params'] = params
            
            
    return fit_fcst_fact, args



"""
UI built using dash to perform basic tasks like uploading user data and setting proper parameters.
"""

from foresee.scripts import main

import base64
import datetime
import io
import os
import sys
import pandas as pd

import flask
import dash
from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_html_components as html
import dash_table


server = flask.Flask(__name__)

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

app = dash.Dash(
					__name__,
					external_stylesheets=external_stylesheets,
					server=server,
					routes_pathname_prefix='/dash/'
				)


app.layout = html.Div([

	### prompt user to provide endog column name ###

	dcc.Markdown('''
	**Provide time series values column name. Default is "y".**
	'''),
	dcc.Input(id="endog-column", type="text", placeholder="time series column", value='y'),
	html.Br(),   

	### prompt user to provide id column name ###

	dcc.Markdown('''
	**If uploading more than one time series, provide an id column name. Default is "id".**
	'''),
	dcc.Input(id="id-column", type="text", placeholder="time series id column", value='id'),
	html.Br(),   

	### prompt user to provide date_stamp column name ###

	dcc.Markdown('''
	**Provide date stamp column name if available. Default is "date_stamp".**
	'''),
	dcc.Input(id="ds-column", type="text", placeholder="Date-Time column", value='date_stamp'),
	html.Br(),   

	### prompt user to provide time series frequency ###

	dcc.Markdown('''
	**Provide time series frequency. Default is 1.**
	'''),
	dcc.Input(id="ts-freq", type="number", placeholder='time series frequency', value=1),
	html.Br(),   

	### prompt user to provide forecast length ###

	dcc.Markdown('''
	**Provide forecast length. Default is 10.**
	'''),
	dcc.Input(id="fcst-length", type="number", placeholder='forecast horizon', value=10),
	html.Br(),   

	### prompt user to provide holdout length ###

	dcc.Markdown('''
	**If comparing model results, provide holdout length for out of sample
	forecast accuracy estimation. Default is 5.**
	'''),
	dcc.Input(id="holdout-length", type="number", placeholder='out of sample holdout len', value=5),
	html.Br(),   

	### display available output formats ###

	dcc.Markdown('''
	**Select result output format.**
	'''),
	dcc.RadioItems(
		id='result-format',
		options=[
			{'label': 'Best Model', 'value': 'best_model'},
			{'label': 'All Models', 'value': 'all_models'},
			{'label': 'All & Best', 'value': 'all_best'}
		],
		value='all_models',
		labelStyle={'display': 'inline-block'}
	),
	html.Br(),

	### display fit-forecast processing method (parallel/sequential) ###

	dcc.Markdown('''
	**Select fit-forecast processing method. Parallel execution with "dask" library.**
	'''),
	dcc.RadioItems(
		id='fit-execution-method',
		options=[
			{'label': 'Parallel', 'value': 'parallel'},
			{'label': 'Sequential', 'value': 'non_parallel'}
		],
		value='non_parallel',
		labelStyle={'display': 'inline-block'}
	),
	html.Br(),

	# display model list

	html.H5('Select or Remove Models'),

	dcc.Checklist(
		id='model-options',
		options=[
			{'label': 'EWM', 'value': 'ewm_model'},
			{'label': 'FFT', 'value': 'fft'},
			{'label': 'Holt Winters', 'value': 'holt_winters'},
			{'label': 'Prophet', 'value': 'prophet'},
			{'label': 'Sarimax', 'value': 'sarimax'}
		],
		value=['ewm_model', 'fft', 'holt_winters', 'prophet', 'sarimax'],
		labelStyle={'display': 'inline-block'}
	),
	html.Br(),

	### display model parameter tuning option ###

	dcc.Markdown('''
	**If tune is selected, model parameters will be optimized using holdout forecast accuracy in "Best Model" and "All & Best" methods.**
	'''),
	dcc.RadioItems(
		id='tune-params',
		options=[
			{'label': 'Tuned Parameters', 'value': 'tune'},
			{'label': 'Default Parameters', 'value': 'default'}
		],
		value='default',
		labelStyle={'display': 'inline-block'}
	),
	html.Hr(),


	### upload file box ###

	dcc.Upload(
		id='upload-data',
		children=html.Div([
			html.H6('Column name (header) is required!!!'),
			'Drag and Drop or ',
			html.A('Select Files'),
		]),
		style={
			'width': '50%',
			'height': '100px',
			'lineHeight': '60px',
			'borderWidth': '1px',
			'borderStyle': 'dashed',
			'borderRadius': '5px',
			'textAlign': 'center',
			'margin': '10px'
		},
		multiple=True
	),

	### display a sample of input data ###

	html.Div(id='output-sample-data', style={'width': '50%'}),
	html.Br(),   

	### display final result dataframe ###

	html.Div(
		id='output-result',
		style={'width': '50%'}
	),
	html.Hr(),

])

### read user input
def read_contents(contents, ds_colname, filename):

	try:
		content_type, content_string = contents.split(',')
		decoded = base64.b64decode(content_string)

		if 'csv' in filename:
			# Assume that the user uploaded a CSV file
			df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))

			if ds_colname in df.columns:
				df[ds_colname] = pd.to_datetime(df[ds_colname])

			return df

		elif 'xls' in filename:
			# Assume that the user uploaded an excel file
			df = pd.read_excel(io.BytesIO(decoded))

			if ds_colname in df.columns:
				df[ds_colname] = pd.to_datetime(df[ds_colname])

			return df

		else:
			# TODO: update this to read txt file
			# Assume that the user uploaded a txt file
			df = pd.read_csv(
				io.StringIO(decoded.decode('utf-8'))
			)
			if ds_colname in df.columns:
				df[ds_colname] = pd.to_datetime(df[ds_colname])

			return df

	except Exception as e:
		return str(e)

### display user input
def parse_contents(contents, ds_colname, filename):

	content = read_contents(contents, ds_colname, filename)

	if type(content) == type(pd.DataFrame()):
		try:
			df_info = [col + ': ' + str(content[col].dtype) for col in content.columns]
			df_info = [x + ' *** ' for x in df_info]
			df_info = '\n'.join([s for s in df_info])

			return html.Div([
				#TODO: fix linebreak
				dcc.Markdown(df_info),

				html.H6('Input data: ' + filename),
				dash_table.DataTable(

					data=content.to_dict('records'),
					columns=[{'name': i, 'id': i} for i in content.columns],
					page_action='none',
					style_table={'height': '200px', 'overflowY': 'auto'},

				),
				html.Hr(),
			])

		except Exception as e:
			return html.H6('Failed to display input data: ' + str(e))


	else:
		return html.H6('Failed to read input data: ' + content)


### display dataframe
def display_dataframe(df, name):

	if type(df) == type(pd.DataFrame()):
		return html.Div([
			html.H5(name),
			dash_table.DataTable(
				data=df.to_dict('records'),
				columns=[{'name': i, 'id': i} for i in df.columns],
				page_action='none',
				style_table={'height': '300px', 'overflowY': 'auto'},              
				export_columns = 'all',
				export_format = 'csv',
			),
			html.Hr(),
		])

	else:
		return html.Div(['run failed with error : ' + str(df)])



### read input file, display a sample
@app.callback(Output('output-sample-data', 'children'),
			  [
				  Input('upload-data', 'contents'),
				  Input('ds-column', 'value'),
			  ],
			  [
				  State('upload-data', 'filename'),
			  ])

def update_output(content_list, ds_colname, filename_list):

	if content_list is not None:
		return parse_contents(content_list[0], ds_colname, filename_list[0])        


### read input file and parameters, display forecast results
@app.callback(Output('output-result', 'children'),
			  [
				  Input('endog-column', 'value'),
				  Input('id-column', 'value'),
				  Input('ds-column', 'value'),
				  Input('ts-freq', 'value'),
				  Input('fcst-length', 'value'),
				  Input('result-format', 'value'),
				  Input('holdout-length', 'value'),
				  Input('model-options', 'value'),
				  Input('upload-data', 'contents'),
				  Input('fit-execution-method', 'value'),
				  Input('tune-params', 'value'),
			  ],
			  [State('upload-data', 'filename')])

def parse_result(
            					endog_colname,
            					gbkey,
            					ds_colname,
            					freq,
            					fcst_length,
            					run_type,
            					holdout_length,
            					model_list,
            					content_list,
            					processing_method,
                       param_type,
            					filename_list,
        				):
    
    if content_list is not None:
        try:
            
            raw_fact = read_contents(content_list[0], ds_colname, filename_list[0])
            raw_fact_cols = raw_fact.columns
            
            if gbkey not in raw_fact_cols:
                gbkey = 'id'
            if ds_colname not in raw_fact_cols:
                ds_colname = 'date_stamp'
            if freq == '':
                freq = 1
            if fcst_length == '':
                fcst_length = 10
            if holdout_length == '':
                holdout_length = 5
            if param_type == 'tune':
                tune = True
            else:
                tune = False
                
            result, fit_result_list = main.collect_result(
                                                                raw_fact,
                                                                endog_colname,
                                                                gbkey,
                                                                ds_colname,
                                                                freq,
                                                                fcst_length,
                                                                run_type,
                                                                holdout_length,
                                                                model_list,
                                                                processing_method,
                                                                tune,
                                                        )
            
            return display_dataframe(result, 'forecast result')
        
        except Exception as e:
            return display_dataframe(str(e), None)

#if __name__ == '__main__':
#    app.run_server(debug=True)
	


"""
test script at dev stage
"""

import warnings
warnings.filterwarnings('ignore')

import os
os.chdir('C:\\Users\\abc_h\\Desktop\\github\\foresee')

# to run in parallel add current directory to path
import sys
sys.path.append(os.getcwd())


import pandas as pd

from foresee.scripts.main import collect_result
from foresee.scripts.utils import read_csv


# sample time-series dataframe with columns(id, date_stamp, y)

ts_df = read_csv('test_data_light.csv')
ts_df['date_stamp'] = pd.to_datetime(ts_df['date_stamp'])
ts_df.head()

# user defind parameters

# time series values column name: required if input dataframe has more than one column

endog_colname = 'y'

# time series frequency
freq = 5

# out of sample forecast length
fcst_length = 10

# available forecasting models
# model_list = ['ewm_model', 'fft', 'holt_winters', 'prophet', 'sarimax']
model_list = ['ewm_model', 'fft', 'holt_winters', 'prophet', 'sarimax']

# avilable run types: 'best_model', 'all_best', 'all_models'
run_type = 'all_best'

# if comparing models (run_type in 'best_model' or 'all_best') then holdout length is required

if run_type == 'all_models':
    holdout_length = None
else:
    holdout_length = 20
    
# fit-forecast computations can be done in parallel for each time series. requires dask library!!!
# for sequential processing set fit_execution_method to 'non_parallel'

fit_execution_method = 'non_parallel'

# if tune is True, search for best parameter. can be slow

tune = True


# since we have two time series in this dataset, time series id column name and date-time column name are required.
gbkey = 'id'
ds_column = 'date_stamp'


result, fit_result_list = collect_result(
                                                    ts_df.copy(),
                                                    endog_colname,
                                                    gbkey,
                                                    ds_column, 
                                                    freq, 
                                                    fcst_length, 
                                                    run_type, 
                                                    holdout_length, 
                                                    model_list,
                                                    fit_execution_method,
                                                    tune,
                                            )



"""
fast fourier transform
"""

import numpy as np
import pandas as pd

# local module
from foresee.models import models_util
from foresee.models import param_optimizer


def reconstruct_signal(
                         n_periods,
                         forecast_len,
                         fft_model,
                         ft_sample_frequencies,
                         fft_terms_for_reconstruction,
                         linear_trend
                      ):
    """

    :param n_periods:
    :param fft_model:
    :param ft_sample_frequencies:
    :param fft_terms_for_reconstruction:
    :param linear_trend:
    :return:
    """
    
    pi = np.pi
    t = np.arange(0, n_periods+forecast_len)
    restored_sig = np.zeros(t.size)
    for i in fft_terms_for_reconstruction:
        ampli = np.absolute(fft_model[i]) / n_periods
        phase = np.angle(
                             fft_model[i],
                             deg = False
                           )
        restored_sig += ampli * np.cos(2 * pi * ft_sample_frequencies[i] * t + phase)
    return restored_sig + linear_trend[0] * t


def fft_fit_forecast(ts, fcst_len, n_harmonics):
    
    ts_len = len(ts)
    
    try:
        t = np.arange(0, ts_len)
        linear_trend = np.polyfit(t, ts, 1)
        training_endog_detrend = ts - linear_trend[0] * t
        fft_model = np.fft.fft(training_endog_detrend)
        indexes = list(range(ts_len))
        
        # sort by amplitude
        indexes.sort(
                        key = lambda i: np.absolute(fft_model[i]) / ts_len,
                        reverse = True
                    )
        fft_terms_for_reconstruction = indexes[:1 + n_harmonics * 2]
        ft_sample_frequencies = np.fft.fftfreq(
                                                    n = ts_len,
                                                    d = 1
                                                 )
        
        fft_fit_forecast = reconstruct_signal(
                                                 n_periods = ts_len,
                                                 forecast_len = fcst_len,
                                                 fft_model = fft_model,
                                                 ft_sample_frequencies = ft_sample_frequencies,
                                                 fft_terms_for_reconstruction = fft_terms_for_reconstruction,
                                                 linear_trend = linear_trend
                                              )
        
        fft_fit_forecast = pd.Series(fft_fit_forecast)
        
        
        fft_fittedvalues = fft_fit_forecast[:-(fcst_len)]
        
        fft_forecast = fft_fit_forecast[-(fcst_len):]
        
        err = None
        
        
    except Exception as e:
        fft_fittedvalues = None
        fft_forecast = None
        err = str(e)
        
    return fft_fittedvalues, fft_forecast, err


def fit_fft(data_dict, freq, fcst_len, model_params, run_type, tune, epsilon):
    
    model = 'fft'
    
    complete_fact = data_dict['complete_fact']
    
    # dataframe to hold fitted values
    fitted_fact = pd.DataFrame()
    fitted_fact['y'] = complete_fact['y']
    fitted_fact['data_split'] = complete_fact['data_split']
    
    # dataframe to hold forecast values
    forecast_fact = pd.DataFrame()
    forecast_fact['y'] = np.full(fcst_len, 0)
    forecast_fact['data_split'] = np.full(fcst_len, 'Forecast')
    
    fit_fcst_fact = pd.concat([fitted_fact, forecast_fact], ignore_index=True)
    
    args = dict()
    
    # no model competition
    if run_type in ['all_models']:
        
        n_harmonics = 5
        fft_fitted_values, fft_forecast, err = fft_fit_forecast(
                                                                    ts = complete_fact['y'],
                                                                    fcst_len = fcst_len,
                                                                    n_harmonics = n_harmonics,
                                                            )
        
        if err is None:
            fit_fcst_fact['fft_forecast'] = fft_fitted_values.append(fft_forecast).values
            
        else:
            fit_fcst_fact['fft_forecast'] = 0
            
        args['err'] = err
        args['n_harmonics'] = n_harmonics
            
    # with model completition            
    if run_type in ['best_model', 'all_best']:
        
        train_fact = data_dict['train_fact']
        test_fact = data_dict['test_fact']
        
        if tune:
            # TODO: add logic when optimization fails
            model_param_space = model_params[model]['n_harmonics']
            n_harmonics, trials = param_optimizer.tune(train_fact, test_fact, fcst_len, model, model_param_space, freq, epsilon)
            args['trials'] = trials
            
        else:
            n_harmonics = 5
            
        training_fitted_values, training_forecast, training_err = fft_fit_forecast(
                                                                                    ts = train_fact['y'],
                                                                                    fcst_len = len(test_fact) ,
                                                                                    n_harmonics = n_harmonics
                                                                                )
        
        complete_fitted_values, complete_forecast, complete_err = fft_fit_forecast(
                                                                                    ts = complete_fact['y'],
                                                                                    fcst_len = fcst_len,
                                                                                    n_harmonics = n_harmonics
                                                                                )
        
        if training_err is None and complete_err is None:
            fft_wfa = models_util.compute_wfa(
                                    y = test_fact['y'].values,
                                    yhat = training_forecast.values,
                                    epsilon = epsilon,
                                )
            fft_fit_fcst = training_fitted_values.append(training_forecast, ignore_index=True).append(complete_forecast, ignore_index=True)
            
            fit_fcst_fact['fft_forecast'] = fft_fit_fcst.values
            fit_fcst_fact['fft_wfa'] = fft_wfa
            
        else:
            fft_wfa = -1
            fit_fcst_fact['fft_forecast'] = 0
            fit_fcst_fact['fft_wfa'] = -1
            
        args['err'] = (training_err, complete_err)
        args['wfa'] = fft_wfa
        args['n_harmonics'] = n_harmonics
        
        
    return fit_fcst_fact, args



# -*- coding: utf-8 -*-
"""
manage input and output of forecasting models
"""

import pandas as pd
import numpy as np
import datetime
import dask
dask.config.set(scheduler='processes')

from foresee.scripts import fitter

def generate_fit_forecast(dict_key, dict_values, model_list, freq, forecast_len, model_params, run_type, tune, epsilon):
    
    fit_result = dict()
    fit_result['ts_id'] = dict_key
    
    for m in model_list:

        f = fitter.fitter(m)

        (
         fit_result[m+'_fit_fcst_df'],
         fit_result[m+'_args']
         ) = f.fit(dict_values, freq, forecast_len, model_params, run_type, tune, epsilon)
    
    return fit_result


# non-parallel fit function
def compose_fit(pre_processed_dict, model_params, param_config, gbkey, run_type, processing_method, tune):
    
    freq = param_config['FREQ']
    forecast_len = param_config['FORECAST_LEN']
    model_list = param_config['MODEL_LIST']
    epsilon = param_config['EPSILON']
    
    # non-parallel fit
    
    if processing_method == 'parallel':
        task_list = list()
        
        for dict_key, dict_values in pre_processed_dict.items():
            fit_task = dask.delayed(generate_fit_forecast)(
                                                               dict_key,
                                                               dict_values,
                                                               model_list,
                                                               freq,
                                                               forecast_len,
                                                               model_params,
                                                               run_type,
                                                               tune,
                                                               epsilon
                                                          )
            task_list.append(fit_task)

        fit_result_list = dask.compute(task_list)[0]
        
    else:
        fit_result_list = list()

        for dict_key, dict_values  in pre_processed_dict.items():
            fit_result = generate_fit_forecast(
                                                    dict_key,
                                                    dict_values,
                                                    model_list,
                                                    freq,
                                                    forecast_len,
                                                    model_params,
                                                    run_type,
                                                    tune,
                                                    epsilon
                                            )
            fit_result_list.append(fit_result)
        
    result = combine_to_dataframe(fit_result_list, model_list, run_type)
        
    return result, fit_result_list


def combine_to_dataframe(fit_result_list, model_list, run_type):
    
    if run_type == 'best_model':

        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:

            bm = result_dict['best_model']
            df = result_dict[bm+'_fit_fcst_df']
            
            df['best_model'] = bm
            df['ts_id'] = result_dict['ts_id']
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
        
    elif run_type == 'all_best':
        
        fit_result_list = _find_best_model(fit_result_list, model_list)
        
        df_list = list()

        for result_dict in fit_result_list:
            bm = result_dict['best_model']
            df = result_dict[bm+'_fit_fcst_df']
            
            for m in model_list:
                mdf = result_dict[m+'_fit_fcst_df']
                df[m+'_forecast'] = mdf[m+'_forecast'].values
                df[m+'_wfa'] = mdf[m+'_wfa'].values

            df['ts_id'] = result_dict['ts_id']
            df['best_model'] = bm
            df_list.append(df)

        result = pd.concat(df_list, axis=0, ignore_index=True)
                
        
    else:
        df_list = list()
        
        for result_dict in fit_result_list:
            df = pd.DataFrame()
            
            for m in model_list:
                mdf = result_dict[m+'_fit_fcst_df']
                df[m+'_forecast'] = mdf[m+'_forecast'].values
                
            df['data_split'] = mdf['data_split'].values
            df['ts_id'] = result_dict['ts_id']
            df_list.append(df)
        
        result = pd.concat(df_list, axis=0, ignore_index=True) 
    
    return result


def _find_best_model(fit_result_list, model_list):
    
    for result_dict in fit_result_list:
        model_wfa_dict = dict()
        
        for m in model_list:
            model_wfa_dict[m] = result_dict[m+'_args']['wfa']
            
        result_dict['best_model'] = [k for k,v in model_wfa_dict.items() if v == max(model_wfa_dict.values())][0]
    
    return fit_result_list


# not in use
def _transform_dataframe_to_dict(raw_fact, gbkey):
    
    data_param_list = list()
    
    if gbkey is None:
        
        data_param_list.append({'ts_id':1, 'df':raw_fact})
    
    else:
        for k,v in raw_fact.groupby(gbkey):
            data_param_list.append({'ts_id':k, 'df':v})
    
    return data_param_list


